
/****** median.c ********************************************************/
/*									*/
/*   Included in this module is a distributed median-finding program	*/
/*   based on the median-of-medians algorithm.				*/
/*									*/
/**************************************************** Shin 7/13/93 ******/

#include <stdio.h>
#include "parti.h"
#include "bisec.h"

#define  MAXPART	512	/* maximum number of partitions */

extern	int	*intlist;

/* find the median of the distributed array */
void
find_median (level, npart, ptr, data, index, proc, midlist, temp)
int    level;			/* dimension of processor partition */
int    npart;			/* current number of partitions */
int    *ptr;			/* pointer of begin of partition */
int    *proc;			/* current processor assignment */
int    *index;			/* indexes of elements in array */
double *data;			/* data array */
Val    *midlist;		/* list of medians */
Val    *temp;			/* temp array for medians */

{
  int    start;			/* start position */
  int    nohigh;		/* number of points higher than median */
  int    psize[MAXPART];	/* sizes of partitions */
  int    gsize[MAXPART];	/* sizes of partitions */
  int    low[MAXPART];		/* index of low */
  int    nolow[MAXPART];	/* number of points lower than median */
  int    high[MAXPART];		/* index of high in array */
  int    mid[MAXPART];		/* index of medians */
  int    midstat[MAXPART];	/* status of median */
  int    nlow[MAXPART];		/* number of points lower than median */
  int    listsize;		/* size of list */
  int    work[MAXPART];		/* working array */
  int    lower;
  int    upper;
  int    done;			/* boolean: global median is found */
  int    found;			/* boolean: local median is found */
  int    i, j, n, no;

  /* set up the proc[], low[], high[], and psize[] */
  for (i = 0; i < npart; i++) {
    /* initialize low and high pointers */
    low[i] = ptr[i];
    high[i] = ptr[i+1] - 1;
    psize[i] = ptr[i+1] - ptr[i];
  }

  /* get the size of partitions */
  for (i = 0; i < npart; i++)
    gsize[i] = psize[i];
  addivec (level, gsize, npart, work);

  /* apply the method median of medians to find 
     the median of the distributed array */
  do {
    /* initialize the median list */
    for (i = npart*PARTI_numnodes; i; ) {
      midlist[--i].val = 0.0;
      midlist[i].index = 0;
    }

    /* set up the  median list */
    for (i = 0, start = PARTI_mynode; i < npart; i++, start += PARTI_numnodes) {
      if (psize[i]) {	/* size > 0 */
        if (high[i] >= low[i]) {
          mid[i] = (high[i] + low[i]) / 2;
          midlist[start].index = 1;
        }
        else    	/* high[i] < low[i] */
          midlist[start].index = 0;
      }
      else {		/* size = 0 */
        mid[i] = ptr[i];
        midlist[start].index = 0;
      }
      midlist[start].val = data[index[mid[i]]];
    }

    /* collect all medians of processors */
    addsvec (level, midlist, PARTI_numnodes * npart, temp);

    /* set up the median list for every partition */
    for (i = 0, start = 0; i < npart; i++, start += PARTI_numnodes) {
      for (j = 0, n=start, no=start; j < PARTI_numnodes; j++, n++) {
        /* check if the size of the partition is equal to NULL */
        if (midlist[n].index) {
          /* size > 0 */
          midlist[no].val = midlist[n].val;
          midlist[no++].index = j;
        }
      }

      /* number of medians in the list */
      listsize = no - start;

      /* sort the median list */
      qsort(&midlist[start], listsize, sizeof(Val), compareval);

      /* median of medians */
      listsize >>= 1;
      midlist[i].val = midlist[start + listsize].val;
      midlist[i].index = midlist[start + listsize].index;

      /* find the position of the median */
      upper = high[i];
      lower = low[i];
      found = FALSE;
      while ((lower <= upper) && (found == FALSE)) {
        /* compart median with data[mid] */
        mid[i] = (upper+lower) / 2;
        midstat[i] = compare(&midlist[i], &data[index[mid[i]]]);

        /* update the low or high */
        switch (midstat[i]) {
        case GREAT:
          lower = mid[i] + 1;
          break;
        case LESS:
          upper = mid[i] - 1;
          break;
        case EQUAL:
          found = TRUE;
          break;
        }
      } 

      /* calculate the number of points that
         are splited by the median */
      if (psize[i] == 0) {/* size = 0 */
        nlow[i] = 0;
        midstat[i] = EQUAL;
      }
      else
      if (midstat[i] > 0) {     /* key >= mid */
        nlow[i] = mid[i] - ptr[i] + 1;
      }
      else {    /* key <= data[mid[i]] */
        nlow[i] = mid[i] - ptr[i];
      }
    }

    /* collect the numbers of points that are greater
       than and less than or equal to the median */
    for (i = 0; i < npart; i++)
      nolow[i] = nlow[i];

    /* count the total number of points that are greater
       than and less than or equal to the median */
    addivec(level, nolow, npart, work);
      
    /* check if the median is the median 
       of the distributed arrays */
    for (i = 0, done = TRUE; i < npart; i++) {
      if (nolow[i] != (gsize[i] / 2)) {
        /* not the median */
        done = FALSE;

        /* update the low or high */
        if (psize[i]) {        /* size > 0 */
          nohigh = gsize[i] - nolow[i];
          switch (midstat[i]) {
          case GREAT:
            if (nolow[i] > nohigh)
              high[i] = mid[i];
            else
              low[i] = mid[i] + 1;
            break;
          case LESS:
            if (nolow[i] > nohigh)
              high[i] = mid[i] - 1;
            else
              low[i] = mid[i];
            break;
          case EQUAL:
            if (nolow[i] > nohigh) {
              high[i] = mid[i] - 1;
              midstat[i] = LESS;
            }
            else {
              low[i] = mid[i] + 1;
              midstat[i] = GREAT;
            }
          }
        }
      }
    }
  } while (done == FALSE);

  /* assign points to processors */
  for (i = 0; i < npart; i++) {
    if (psize[i] == 0)
      /* no points in this partition */
      continue;

    /* points that are less than data[mid[i]]
       are assigned to processor 2i */
    for (j = ptr[i]; j < ptr[i+1]; j++) 
      proc[index[j]] <<= 1;

    /* point data[mid[i]] */
    if (midstat[i] <= 0)
      proc[index[mid[i]]]++;

    /* points that are greater than the median */
    /* are assigned to processor 2i + 1 */
    for (j = mid[i] + 1; j < ptr[i + 1]; j++)
      proc[index[j]]++;
  }

  /* partition arrays */
  for (intlist = proc, i = npart - 1; i >= 0; i--) {
    qsort(&index[ptr[i]], psize[i], sizeof(int), compareint);
    ptr[2*i] = ptr[i];    /* ith partition is splited */
    ptr[2*i+1] = ptr[i] + nlow[i];
  }
}

/* find the median of the weighted distributed array */
void
find_median_weight (level, npart, ptr, data, weight, index, proc, midlist, midtemp)
int    level;			/* level of binary disection */
int    npart;			/* current number of partitions */
int    *ptr;			/* pointer of begin of partition */
int    *index;			/* indexes of elements in array */
int    *proc;			/* current processor assignment */
int    *weight;			/* wight of each element */
double *data;			/* data array */
Val    *midlist;		/* list of medians */
Val    *midtemp;		/* temp array for medians */

{
  int    start;			/* start position */
  int    nohigh;		/* number of points higher than median */
  int    psize[MAXPART];	/* sizes of partitions */
  int    gsize[MAXPART];	/* sizes of partitions */
  int    low[MAXPART];		/* index of low */
  int    nolow[MAXPART];	/* number of points lower than median */
  int    high[MAXPART];		/* index of high in array */
  int    mid[MAXPART];		/* index of medians */
  int    midstat[MAXPART];	/* status of median */
  int    nlow[MAXPART];		/* number of points lower than median */
  int    done[MAXPART];		/* list of status */
  int    work[MAXPART];
  int	 midload;		/* load of mid */
  int	 halfsize;		/* half of total size */
  int    listsize;		/* size of list */
  int	 tempmid;
  int    lower;
  int    upper;
  int    alldone;		/* boolean */
  int    found;			/* boolean: median is found */
  int    i, j, k, n, no;

  /* set up the proc[], low[], high[], and psize[] */
  for (i = 0; i < npart; i++) {
    /* initialize low and high pointers */
    low[i] = ptr[i];
    high[i] = ptr[i + 1] - 1;
    psize[i] = ptr[i+1] - ptr[i];
  }

  /* get the size of partitions */
  for (i = 0; i < npart; i++)
    gsize[i] = weight[ptr[i+1]] - weight[ptr[i]];
  addivec (level, gsize, npart, work);

  /* initial done[] array */
  for (i = 0; i < npart; i++) {
    if (gsize[i])
      done[i] = FALSE;
    else		/* empty */
      done[i] = TRUE;
  }

  /* apply the method median of medians to find 
     the median of the distributed array */
  do {
    /* initialize the median list */
    for (i = npart*PARTI_numnodes; i; ) {
      midlist[--i].val = 0.0;
      midlist[i].index = 0;
    }

    /* find the media of local arrays */
    for (i = 0, start = PARTI_mynode; i < npart; i++, start+=PARTI_numnodes) {
      /* check if median is already found */
      if (done[i]) continue;

      if (psize[i]) {	/* size > 0 */
        if (high[i] >= low[i]) {
          mid[i] = (high[i] + low[i]) / 2;
	  if ((midload = weight[mid[i]+1] - weight[mid[i]]))
            midlist[start].index = midload;
	  else		/* load of mid is 0 */
	    midlist[start].index = 1;
        }
        else		/* high[i] < low[i] */
          midlist[start].index = 0;
      }
      else {		/* size = 0 */
        mid[i] = ptr[i];
        midlist[start].index = 0;
      }
      midlist[start].val = data[index[mid[i]]];
    }

    /* collect all medians of processors */
    addsvec (level, midlist, PARTI_numnodes * npart, midtemp);

    /* set up the median list for every partition */
    for (i = 0, start = 0; i < npart; i++, start += PARTI_numnodes) {
      /* check if median is already found */
      if (done[i]) continue;

      for (j = 0, n=start, no=start; j < PARTI_numnodes; j++, n++) {
        /* check if the size of the partition is equal to NULL */
        if (midlist[n].index) {
          /* size > 0 */
          midtemp[no].val = midlist[n].val;
          midtemp[no++].index = j;
        }
      }

      /* number of medians in the list */
      listsize = no - start;

      /* check if the median list is empty */
      if (listsize == 0) continue;

      /* sort the median list */
      qsort(&midtemp[start], listsize, sizeof(Val), compareval);

      /* median of medians */
      listsize >>= 1;
      midtemp[i].val = midtemp[start + listsize].val;
      midtemp[i].index = midtemp[start + listsize].index;

      /* find the position of the median */
      upper = high[i];
      lower = low[i];
      found = FALSE;
      while ((done[i] == FALSE) && (lower <= upper) && (found == FALSE)) {
        /* compart median with data[mid] */
        mid[i] = (upper+lower) / 2;
        midstat[i] = compare(&midtemp[i], &data[index[mid[i]]]);

        /* update the low or high */
        switch (midstat[i]) {
        case GREAT:
          lower = mid[i] + 1;
          break;
        case LESS:
          upper = mid[i] - 1;
          break;
        case EQUAL:
          found = TRUE;
          break;
        }
      } 

      /* calculate the number of points that
         are splited by the median */
      if (psize[i] == 0) {/* size = 0 */
        nlow[i] = 0;
	midstat[i] = EQUAL;
      }
      else if (midstat[i] > 0) {   /* key > mid */
        nlow[i] = weight[mid[i]+1] - weight[ptr[i]];
      }
      else {  /* key <= data[mid[i]] */
        nlow[i] = weight[mid[i]] - weight[ptr[i]];
      }
    }

    /* collect the numbers of points that are greater
       than and less than or equal to the median */
    for (i = 0; i < npart; i++)
      nolow[i] = nlow[i];

    /* count the total number of points that are greater
       than and less than or equal to the median */
    /* gisum(nolow, npart, work); */
    addivec(level, nolow, npart, work);

    /* check if the median is the median of the distributed arrays */
    /* i.e. nolow <= gsize/2 < nolow+load[mid] */
    for (i=0, alldone=TRUE, start=0; i < npart; i++, start+=PARTI_numnodes) {
      if (done[i]) continue;	/* median is already found */
      halfsize = gsize[i] >> 1;

      if (nolow[i] > halfsize) {
	/* median is in the lower half */
	switch (midstat[i]) {
	case GREAT:
	  high[i] = mid[i];
	  break;
	case LESS:
	  high[i] = mid[i] - 1;
	  break;
	case EQUAL:
	  high[i] = mid[i] - 1;
	  midstat[i] = LESS;
	  break;
	}

	/* go on searching the lower half */
	alldone = FALSE;
	continue;
      }

      tempmid = start + midtemp[i].index;
      if (halfsize >= (nolow[i] + midlist[tempmid].index)) {
	/* median is in the upper half */
	switch (midstat[i]) {
	case GREAT:
	  low[i] = mid[i] + 1;
	  break;
	case LESS:
	  low[i] = mid[i];
	  break;
	case EQUAL:
	  low[i] = mid[i] + 1;
	  midstat[i] = GREAT;
	  break;
	}

	/* go on searching the upper half */
	alldone = FALSE;
	continue;
      }

#ifdef	DEBUG
      if (PARTI_mynode == 0) {
	int lowdiff = halfsize-nolow[i];
	int highdiff = nolow[i]+midlist[tempmid].index-halfsize;

	midload = midlist[tempmid].index;
	printf ("part = %d gsize = %d halfsize = %d ", i, gsize[i], halfsize);
	printf ("load = %4d lowdiff = %4d highdiff = %4d\n", midload, lowdiff, highdiff);
      }
#endif
      /* median is found */
      done[i] = TRUE;

      /* decide halfsize is closer to nolow or (nolow + load[mid]) */
      if ((halfsize-nolow[i]) > (nolow[i]+midlist[tempmid].index-halfsize)) {
	if (PARTI_mynode == midtemp[i].index)
	  midstat[i] = GREAT;
      }
      else {
	if (PARTI_mynode == midtemp[i].index)
	  midstat[i] = LESS;
      }
    }
  } while (alldone == FALSE);

  /* assign points to processors */
  for (i = 0; i < npart; i++) {
    if (psize[i] == 0)
      /* no points in this partition */
      continue;

    /* points that are less than data[mid[i]]
       are assigned to processor 2i */
    for (j = ptr[i]; j < ptr[i+1]; j++) 
      proc[index[j]] <<= 1;

    /* point data[mid[i]] */
    if (midstat[i] <= 0)
      proc[index[mid[i]]]++;

    /* points that are greater than the median */
    /* are assigned to processor 2i + 1 */
    for (j = mid[i] + 1; j < ptr[i + 1]; j++) 
      proc[index[j]]++;
  }

  /* partition arrays */
  for (intlist = proc, i = npart - 1; i >= 0; i--) {
    qsort(&index[ptr[i]], psize[i], sizeof(int), compareint);
    ptr[2*i] = ptr[i];  /* ith partition is splited */
    if (midstat[i] <= 0)
      ptr[2*i+1] = mid[i];
    else
      ptr[2*i+1] = mid[i] + 1;
  }
}


